unit VennDiagram;
(*
   Komponentti venn-kaavioiden esittmiseen ja animointiin
   KESKENERINEN.
   Miika Nurminen, 25.4.2004.

   Luokat:
    ColorMask: sislt snnt, joiden mukaan kaavio
    piirretn. snnt ovat loogisia lausekkeita
    konjunktiivisessa? normaalimuodossa, esim
     (a1 AND a2) OR (a1 AND (NOT a3)) OR (...)

    Yksi colormask-olio sislt yhden AND-sivulauseen.
    colormaskin taulukon alkiot vastaavat piirrettvi
    olioita. snnt kuvataan predikaateilla
      pN = ei vli, pT = tosi, pF = eptosi

    checkmask-metodilla kysytn, onko sivulause voimassa
    nykyisill ehdoilla.

    TVennDiagram vastaa kaavion piirrosta. Kaavioita
    voidaan animoida (tll hetkell vain yksinertainen
    pyritys) asettamalla Venn-kaavion timer-ominaisuus.
*)
// tee rajapintaa kyttv versio ja peri neli siit
// mahdollisesti mys ellipsiversio?

interface

uses
  SysUtils, Classes, Controls, ExtCtrls, Graphics,
  contnrs, types;

type
  TPredicate = (pN,pT,pF);

  TColorMask = class
  private
    mask:array of TPredicate;
    color:TColor;
  public
    constructor Create(acolor:TColor); overload;
    constructor Create(acolor:TColor;amask:array of TPredicate); overload;
    destructor Destroy; override;
    procedure SetSize(asize:integer);
//    class function GetSize:integer;
    procedure SetMask(amask:array of TPredicate); overload;
    procedure SetMask(num:integer;val:TPredicate); overload;
    function CheckMask(achk:array of boolean):boolean;
  end;


  IBoundsObject = interface
    function Sisalla(x,y:integer):boolean;
  end;

  TVennDiagram = class(TPaintBox)
  private
    FMasks:TObjectList;
    FSetCount: integer;
    ControlArray:array of TRect;
    tempc:array of boolean;
    FTimer: TTimer;
    ofs:Real;
    onOldTimer:TNotifyEvent;
    procedure onTimer(Sender:TObject);
    procedure SetSetCount(const Value: integer);
    function GetPredicates(m: integer): TColorMask;
    function Sisalla(x,y:integer;r:TRect):boolean;
    function GetPredicateCount: integer;
    procedure SetTimer(const Value: TTimer);
  protected
    procedure Paint; override;
  public
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure AddPredicate(color:TColor); overload;
    procedure AddPredicate(color:TColor;amask:array of TPredicate); overload;
    property Predicates[m:integer]: TColorMask read GetPredicates;
    property PredicateCount: integer read GetPredicateCount;
  published
    property SetCount:integer read FSetCount write SetSetCount;
    property Timer:TTimer read FTimer write SetTimer;
  end;

procedure Register;

implementation

//var masksize:integer;

procedure Register;
begin
  RegisterComponents('GKOC', [TVennDiagram]);
end;

{ TColorMask }

function TColorMask.CheckMask(achk: array of boolean): boolean;
var i:integer;valid:boolean;
begin
  Result:=false;
  if (Length(mask)=0) then Exit;
  Valid:=false;
  for i:=Low(mask) to High(mask) do begin
   if (mask[i]=pT) and (achk[i]=false) then Exit;
   if (mask[i]=pF) and (achk[i]=true) then Exit;
   if (mask[i]<>pN) then valid:=true;
  end;
  if not Valid then Exit;
  // jos kaikki ehdot ovat negatiivisia, ei piirret
  // mitn (muuten koko ruutu piirrettisiin)
  Result:=true;
end;

constructor TColorMask.Create(acolor: TColor);
begin
  SetLength(mask,0);
  Color:=acolor;
end;

constructor TColorMask.Create(acolor: TColor; amask: array of TPredicate);
begin
  self.Create(acolor);
  SetMask(amask);
end;

destructor TColorMask.Destroy;
begin
  mask:=nil;
  inherited;
end;


procedure TColorMask.SetMask(num: integer; val: TPredicate);
begin
  mask[num]:=val;
end;

procedure TColorMask.SetMask(amask: array of TPredicate);
var i:integer;
begin
  SetSize(Length(amask));
  if Length(amask)=0 then Exit;
  for i:=low(amask) to high(amask) do begin
    mask[i]:=amask[i];
  end;
end;

procedure TColorMask.SetSize(asize: integer);
begin
  if (Length(mask)<>asize) then SetLength(mask,asize);
end;

{ TVennDiagram }



procedure TVennDiagram.AddPredicate(color: TColor;
  amask: array of TPredicate);
begin
  FMasks.Add(TColorMask.Create(color,amask));
end;

procedure TVennDiagram.AddPredicate(color: TColor);
begin
  FMasks.Add(TColorMask.Create(color));
end;

constructor TVennDiagram.Create(AOwner: TComponent);
begin
  inherited;
  FMasks:=TObjectList.Create(true);
  FSetCount:=0;
end;

destructor TVennDiagram.Destroy;
begin
  FMasks.Free;
  ControlArray:=nil;
  tempc:=nil;
  inherited;
end;

function TVennDiagram.GetPredicateCount: integer;
begin
  Result:= FMasks.Count;
end;

function TVennDiagram.GetPredicates(m: integer): TColorMask;
begin
  Result := FMasks[m] as TColorMask;
end;

procedure TVennDiagram.Paint;
var i,j,k:integer;
m:TColorMask;
a,b:real;x,y,xt,yt,xs,ys:integer;c:real;
begin
  inherited;
  if SetCount=0 then Exit;

  a:=2*pi / SetCount;
  b:=ofs;
  c:=0.7;
  xt:=Width div 2;
  yt:=Height div 2;
  xs:=Width div 4;
  ys:=Height div 4;
 Canvas.Brush.Color:=clBlack;
  k:=0;
  while (b<(2*pi+ofs)) and (k<SetCount) do begin
    x:=round(c*cos(b)*xs)+xt;
    y:=round(c*sin(b)*ys)+yt;

    with ControlArray[k] do begin
      Left:=x-xs;
      Top:=y-ys;
      Right:=x+xs;
      Bottom:=y+ys;
    end;
    b:=b+a;

    inc(k);
  end;

  if PredicateCount>0 then
    for i:=0 to Width do for j:=0 to Height do begin
      for k:=0 to SetCount-1 do begin
        tempc[k]:=Sisalla(i,j,ControlArray[k]);
      end;
      for k:=0 to PredicateCount-1 do begin
        m := Predicates[k];
        if m.CheckMask(tempc) and (((i+j+k) mod PredicateCount) = 0) then
          Canvas.Pixels[i,j]:=m.color;
      end;
    end;

  for k:=0 to SetCount-1 do begin
   Canvas.FrameRect(ControlArray[k]);
  end;
end;



procedure TVennDiagram.SetSetCount(const Value: integer);
var i:integer;
begin
  if (Value<0) then Exit;

  SetLength(tempc,Value);
  FSetCount := Value;
  SetLength(ControlArray,Value);
  if PredicateCount>0 then
    for i:=0 to PredicateCount-1 do
      Predicates[i].SetSize(Value);
  Invalidate;
end;




procedure TVennDiagram.SetTimer(const Value: TTimer);
begin
  if (Value=Ftimer) then Exit;
  if Assigned(FTimer) then begin
    Ftimer.RemoveFreeNotification(self);
    Ftimer.OnTimer:=onOldTimer;
  end;
  FTimer := Value;
  if Assigned(FTimer) then begin
    onOldTimer:=FTimer.onTimer;
    FTimer.OnTimer:=onTimer;
    FTimer.FreeNotification(self);
  end;
end;

function TVennDiagram.Sisalla(x, y: integer; r: TRect): boolean;
begin
  Result:=false;
  if x<r.Left then Exit;
  if y<r.Top then Exit;
  if x>=r.Right then Exit;
  if y>=r.Bottom then Exit;
  Result := true;
end;

procedure TVennDiagram.onTimer(Sender: TObject);
begin
  if csDesigning in ComponentState then Exit;
  if Assigned(onOldTimer) then onOldTimer(Sender);
  ofs:=ofs+0.05;
  if ofs>(2*pi) then ofs:=0;
  Invalidate;
end;

end.
